/*---------------------------------------------------------------------------*
  Project:  SoftKeyboard
  File:     KeyLayoutConverter.cpp

  Copyright 2005 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: KeyLayoutConverter.cpp,v $
  Revision 1.1  2005/06/14 11:30:18  hatamoto_minoru
  Initial registration

  $NoKeywords: $
 *---------------------------------------------------------------------------*/

// KeyLayoutConverter.cpp : Defines the console application entry point.
//

#include "stdafx.h"

using namespace std;

struct TGAHeader
{
	u8   id_len; // Number of Characters in Identification Field. 
	u8	cmap_type; // Color Map Type. 
	u8	img_type; // Image Type Code
	u8   cmap_spec[5]; // Color Map Specification
	u16	ox;				// X Origin of Image in dot
	u16	oy;				// X Origin of Image in dot
	u16	width;          // width in dot
	u16	height;			// height in dot
	u8	pixsize;	    // Image Pixel Size in bit
	u8	id_field;		// Image Identification Field
};


struct Rect
{
    u8 x, y, width, height;
};


typedef struct KbdButton
{
    Rect rect; // button rectangle
    s16  groupoffset; // layout index offset for elements within a group
    u16 id; // button id
} KbdButton;


/// onfirms whether a color in one place is the same
bool EqualInColor( const u32* first_begin,
					const u32* first_end,
					const u32* second_begin,
					u32 color )
{
	for( int i=0; i<distance(first_begin, first_end); i++ )
	{
		// Is one the color?
		if( first_begin[i] == color 
			|| second_begin[i] == color )
		{
			// Is the content different?
			if( first_begin[i] != second_begin[i] )
			{
				return false;
			}
		}
	}

	return true;

}


// A8 R8 G8 B8 -> H16 B8 S8
struct ARGB_to_HBS
    : unary_function<u32,u32>
{
    u32 operator()( u32 argb ) const
      {
          double r, g, b;

          r = (argb & 0x00FF0000) >> 16;
          g = (argb & 0x0000FF00) >>  8;
          b = (argb & 0x000000FF) >>  0;

          double rgb[3] = {r,g,b};

          double min = *min_element(rgb,rgb+3);
          double max = *max_element(rgb,rgb+3);

          // hue calculations
          double hue;

          if( max == 0 )
          {
              hue = 0;
          }
          else if (g == max)
          {
              hue= (b - r) / ( max - min ) * 60 + 120;
          }
          else if (b == max)
          {
              hue= (r - g) / ( max - min ) * 60 + 240;
          }
          else if (g<b)
          {
              hue= (g - b) / ( max - min ) * 60 + 360;
          }
          else
          {
              hue= (g - b) / ( max - min ) * 60;
          }

          // saturation calculation

          double saturation = ( max == 0 ? 0 : ( max - min ) / max * 100 );

          // luminance calculation

          double brightness = max;

          return (u16)hue << 16 | (u8)brightness << 8 | (u8) saturation;
          
      }
};






Rect GetOneRect( u32* pix,
                 int width,
                 int height,
                 int start_y,
                 u32 color )
{
    u32* first;
	u32* second;
    int y;

    // ....oo...
    // ....oo... 
    // ...ooo... search for this column (non-matching content), and return a rectangle above it
    // ...ooo...
    for( y = start_y+1; y<height; y++ )
    {
        first  = pix + width * (y-1);
        second = pix + width * y;
        
        if( ! EqualInColor( first, first + width, second, color ) ) break;
    }

    u32* lastline = pix + width*(y-1);
    
    Rect rect;

    rect.x = (u8)distance( lastline, find(lastline,lastline+width, color) );
    rect.y = (u8)start_y;
    rect.width = (u8)count( lastline, lastline + width, color );
    rect.height = (u8)( y - start_y );

    return rect;
    
}
                    
                    


void AppendRects( u32* pix,
                  int width,
                  int height,
                  u32 color,
                  list<Rect>* rects
    )
{
    for( int y=0; y<height; y++ )
    {
        if( count( pix + y*width, pix + (y+1)*width, color ) > 0 )
        {
            Rect r = GetOneRect(pix, width, height, y, color );
            rects->push_back( r );
            y += r.height;
        }
    }
}


int _tmain(int argc, _TCHAR* argv[])
{

	if( argc != 3 )
	{
		cerr << "usage: KeyLayoutConverter <infile> <outfile>" << endl;
		return 1;
	}

	char* infilename = argv[1];
	char* outfilename = argv[2];

	TGAHeader header;
	
	fstream infile( infilename, ios::in | ios::binary );

	if( infile.fail() )
	{
		cerr << "failed to open file : " << infilename << endl;
		return 1;
	}

	infile.read( (char*)&header, sizeof(header) );

  if( header.pixsize != 32 )
  {
		cerr << "infile must be 32bit tga file" << endl;
		return 1;
  }

	u32* pixbuf = new u32[header.width * header.height]; 
 
	// tga file is DTLR. This is read in TDLR.
    for( int i=0; i<header.height; i++ )
    {
		infile.read( (char*)(&pixbuf[header.width*(header.height-i-1)]), sizeof(u32)*header.width );
	}

    // convert ARGB to HBS

    transform( pixbuf, pixbuf + header.width*header.height, pixbuf,
               ARGB_to_HBS() );

	map< u32, list<Rect>* > regions;

    // first, list out all the colors
	for( int i=0; i<header.width*header.height; i++ )
	{
        // if there are no colors, append an empty list
        if( regions.count( pixbuf[i] ) == 0 )
        {
            regions[pixbuf[i]] = new list<Rect>();
        }
	}

	regions.erase( 0 ); // delete the blacks

    map< u32, list<Rect>* >::iterator it;

    for( it=regions.begin(); it != regions.end(); ++it )
    {
        AppendRects( pixbuf, 
			         header.width,
					 header.height,
					 (*it).first, 
					 (*it).second );
    }


	vector<KbdButton> buttons;
	buttons.reserve( regions.size() );

	int id = 1;

	for( it=regions.begin(); it != regions.end(); ++it )
	{
		list<Rect>* rects = (*it).second;

		list<Rect>::iterator rectitr;
		
		for( rectitr=rects->begin(); rectitr != rects->end(); ++rectitr )
		{
			KbdButton element;
			
			element.rect = *rectitr;

			if( rects->size() == 1 )
			{
				element.groupoffset = 0;
			}
			else
			{
				if( distance( rects->begin(), rectitr ) == rects->size()-1 ) // if the last element
				{
					element.groupoffset = (s16)( 1 - s32(rects->size()) );
				}
				else
				{
					element.groupoffset = 1;
				}
			}

			element.id = id;

			buttons.push_back( element );

		}

		id ++;
	}


	fstream outfile( outfilename, ios::out | ios::binary );

	if( outfile.fail() )
	{
		cerr << "failed to open outfile : " << outfilename << endl;
		return 1;
	}

	outfile.write( (char*)( &buttons[0] ), streamsize(sizeof(KbdButton)*buttons.size()) );

// delete has not been used at all. Go for it, OS.

	return 0;
}

