/*---------------------------------------------------------------------------*
  Project:  SoftKeyboard
  File:     kbd.c

  Copyright 2005 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: kbd.c,v $
  Revision 1.5  2006/03/01 19:57:00  tahara_yasuyuki
  Support key top design change
  
  Revision 1.4  2005/07/07 11:20:57  tahara_yasuyuki
  Comment revisions

  Revision 1.3  2005/07/07 06:58:49  tahara_yasuyuki
  Support for euro keyboard

  Revision 1.2  2005/07/04 01:10:58  hatamoto_minoru
  Comment revisions

  Revision 1.1  2005/06/14 11:30:05  hatamoto_minoru
  Initial registration

  $NoKeywords: $
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
  * As a result of the keytop drawing, non-special characters must be rectangles.
  * Be aware that there are three numerical series involved:
  character code (UTF 16), scan code (key numbers) and the layout index (button coordinates array index).
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include <string.h>
#include <wstring.h>
#include <nnsys/g2d/g2d_CharCanvas.h>
#include "kbd_autoRepeat.h"
#include "kbd_data.h"
#include "kbd_input.h"
#include "kbd_font.h"
#include "kbd.h"

#define MIN(a, b)               (((a) < (b)) ? (a) : (b))
#define MAX(a, b)               (((a) > (b)) ? (a) : (b))
#define CHARCOLOR 7

#define KBD_WIDTH 256 // [pix]
#define KBD_HEIGHT 96 // [pix]
//#define KBD_WIDTH 200 // [pix]
//#define KBD_HEIGHT 88 // [pix]

#define RSRC_SIZE( sym ) (u32)(( sym##_end - sym ))

// SCANCODE
enum
{
    SCANCODE_DELETE = 1,
    SCANCODE_RETURN = 2, 
    SCANCODE_SPACE  = 3
};


// SCANCODE (Japanese-specific)
enum
{
    SCANCODE_HIRAGANA = 4,
    SCANCODE_KATAKANA,
    SCANCODE_DAKUTEN,
    SCANCODE_HANDANKU,
    SCANCODE_SMALL
};


// SCANCODE (ASCII, EUR-specific)
enum
{
    SCANCODE_SHIFT = 4,
    SCANCODE_CAPS
};


// local special code for created voiced and semi-voiced consonant diacritic characters and small characters
enum
{
    CHARCODE_DAKUTEN = L'\a', // bell replacement
    CHARCODE_HANDAKU = L'\t', // tab replacement
    CHARCODE_SMALL = L'\v' // vertical tab replacement
};


struct KbdControl;

typedef void (*KeyFunc)( struct KbdControl* ctrl );



typedef struct KbdConfig
{
    const KBDButton*      layout;     // button layout
    void*           charimg;    // keytop character image
    KeyFunc         key_func; // function called when key is pressed
    u16             specialkey_num; // number of special keys
    u16             keynum;     // number of keys (including special keys)
    
} KbdConfig;



// keyboard internal state
typedef struct KbdControl
{
    const KBDButton*       pressed; // button at touch-on
    const KBDButton*       current; // touched button
    const KbdConfig* conf;   // pointer to keyboard settings
    KBDDragDropCallback dragdropcallback; // Function for drag & drop. if none, NULL
    void*            dragdropparam; // drag & drop callback parameters
    KBDSoundCallback soundcallback; //-- Function for playing sound. if none, NULL
    void*            soundparam; //-- sound callback parameters
    u16*             keymap; // conversion chart from scan code to character code
    KBDAutoRepeat    repeat;
    u16              scancode; // input key (key number)
    u16              input;    // input key (character code)
    u16              dragging; // dragged character
    u16              highlight;  // Key number for highlighted key. used for SHIFT and CAPS LOCK
    KBDKind          kind;

    const KBDButton*       cursor; // button under the cursor
    BOOL             cursor_input; // flag for cursor input

    BOOL             bootmenu_mode; // Is a dummy button to be used? The dummy button can be used to accept entries in name input.
    KBDExtTpEventCallback ext_tpevent_callback;  //-- Function at additional touch events. if none, NULL
    void*            ext_tpevent_param;  //-- additional touch event callback parameters
    const KBDButton* ext_tpevent;   //-- additional touch event list pointer
    u16              ext_tpevent_num;  //-- number of additional touch event registries
    u16              pad;  //-- padding

} KbdControl;






#define RECT_L( rect ) ( (rect).x )
#define RECT_R( rect ) ( (rect).x + (rect).width )
#define RECT_T( rect ) ( (rect).y )
#define RECT_B( rect ) ( (rect).y + (rect).height )


static const int c_KBD_X = 40;
static const int c_KBD_Y = 56;

static u8         sKbdImg[KBD_WIDTH*KBD_HEIGHT/2 + sizeof(GXCharFmt16) ]; // place that drawn by the keytop and the space character in the ncg data
static u8         sIcgBak[KBD_WIDTH*KBD_HEIGHT/2 + sizeof(GXCharFmt16) ]; // use this to reverse the color of pressed keys
static int        sPosX ,sPosY; // the keyboard's upper-left coordinates (for conversion between screen and keyboard touch coordinates)
static KbdControl sControl;     // keyboard state
static BOOL       s_Initialized;


static KBDButton s_dummybuttons[] =
{
    { { 140 - c_KBD_X + 4, 144 - c_KBD_Y + 1, 100 - 8, 16 - 3 }, 0, 0 }, // Deletes a border pixel by pixel. used to make the cursor more visible
    { {  40 - c_KBD_X + 4, 144 - c_KBD_Y + 1, 100 - 8, 16 - 3 }, 0, 0 },
}; // Dummy button to determine if the cursor has been moved. Use for name comment input.


static const u16* s_all_map[] =
{
    L"ĂƂ͂ЂӂւقACEGIJLNPRTVXZ\^`cegnqtwz",
    L"@@@@@Âłǂ΂тԂׂځ@@@@@@@@KMOQSUWY[]_adfhorux{@@@@",
    L"@@@@@@@@@@@@@@@@@@@@ς҂Ղ؂ہ@@@@@@@@@@@@@@@@@@@@@@@@psvy|@@@@",
    L"@@@@@@@@@@@@@@@@@@@@BDFH@@@@@@@@@@b@@@@@@@",
};

static const u16* s_dakuten_map[2];
static const u16* s_handaku_map[2];
static const u16* s_small_map[2];


// static functions
static void PlayInputSE( u16 scancode );
static u16  kbd_get_input_internal( KbdControl* ctrl, u16 scancodevoid );
static void UpdateDragging( KbdControl* ctrl );
static void AsciiEuroCommonKBDFunc( KbdControl * ctrl, const u8* map_binaries[] );
static u16 ConvertDakutenEtc( const u16* map[], u16 c  );
static const KBDButton* SearchButton( const KbdControl* ctrl,
                                      int x, ///< [in] X in absolute coordinates
                                      int y  ///< [in] Y in absolute coordinates
    );
static const KBDButton* MoveCursor( int key, const KBDButton* current, const KBDButton* layout );
static void RelocateCursor( KbdControl* ctrl );
static void SignKBDInit( KbdControl* ctrl );
static void SignKBDFunc( KbdControl* /* ctrl */ );

static void AddGridRectangle( u8* gridstart,
                              u32 add_value,
                              s32 left, s32 top, s32 right, s32 bottom );

static void AddBGCharRectangle( u8* bgchardata,
                                s32 data_width,
                                u32 add_value,
                                s32 left,
                                s32 top,
                                s32 right,
                                s32 bottom  );

static void AsciiKBDInit( KbdControl* ctrl );
static void JapaneseKBDInit( KbdControl* ctrl );
static void EuroKBDInit( KbdControl* ctrl );
static void SignKBDInit( KbdControl* ctrl );
static void PictoKBDInit( KbdControl* ctrl );

static void AsciiKBDFunc( KbdControl* ctrl );
static void JapaneseKBDFunc( KbdControl* ctrl );
static void EuroKBDFunc( KbdControl* ctrl );
static void SignKBDFunc( KbdControl* ctrl );
static void PictoKBDFunc( KbdControl* ctrl );

static void KBDi_SoundCallback( void* param, int sound );

// keyboard settings data
static KbdConfig sConfig[KBD_KIND_NUM] =
{
    {
		(KBDButton *)ascii_kbd,
		fs_chat_keyboard_ncg,
		AsciiKBDFunc,
		5,
		0, // 0 is entered in advance and then the value is calculated at runtime
    }, // ASCII
	
    {
		(KBDButton *)jp_kbd,
		fs_chat_keyboard_jp_ncg,
		JapaneseKBDFunc,
		8,
		0,
    }, // JAPANESE

    {
		(KBDButton *)picto_kbd, // same layout as for PICTO
		fs_chat_keyboard_euro_ncg,
		EuroKBDFunc,
		3,
		0,
    }, // EUROPE

    {
		(KBDButton *)picto_kbd, // same layout as for PICTO
		fs_chat_keyboard_picto_ncg,
		PictoKBDFunc,
		3,
		0,
    }, // SIGN

    {
		(KBDButton *)picto_kbd,
		fs_chat_keyboard_picto_ncg,
		PictoKBDFunc,
		3,
		0,
    }, // PICTURE

};





/*---------------------------------------------------------------------------*

  Returns the button layout index for ordinary key headers.
  Uses keytop drawing.

 *---------------------------------------------------------------------------*/
static u16 getNormalButtonStart( const KbdConfig* conf )
{
    u16 i;

    for ( i = 0; i < conf->keynum ; i++ )
    {
        if( conf->layout[i].id == conf->specialkey_num + 1) return i; // IDs begin with 1, so +1
    }

    SDK_ASSERT( FALSE );
    return 0;

}



///////////////////////////////////////////////////////////////////////////////
///
/// Gets the button pointer from coordinates.
/// Returns NULL when no button has been touched.
///
///////////////////////////////////////////////////////////////////////////////
static const KBDButton* SearchButton( const KbdControl* ctrl,
                                      int x, ///< [in] X in absolute coordinates
                                      int y  ///< [in] Y in absolute coordinates
    )
{
    const KBDButton* layout;
    int i, keynum;

    layout = ctrl->conf->layout;
    keynum = ctrl->conf->keynum;

    for( i=0; i<keynum; i++ )
    {

        if( ( RECT_T( layout[i].rect ) + sPosY ) <= y && y < ( RECT_B( layout[i].rect ) + sPosY )
            && ( RECT_L( layout[i].rect ) + sPosX ) <= x && x < ( RECT_R( layout[i].rect ) ) + sPosX )
        {
            return &layout[i];
        }

    }

    if( ctrl->bootmenu_mode )
    {
        layout = s_dummybuttons;

        for( i = 0; i < KBD_ARRAY_SIZE(s_dummybuttons); i++ )
        {
            if( ( RECT_T( layout[i].rect ) + sPosY ) <= y && y < ( RECT_B( layout[i].rect ) + sPosY )
                && ( RECT_L( layout[i].rect ) + sPosX ) <= x && x < ( RECT_R( layout[i].rect ) ) + sPosX )
            {
                return &layout[i];
            }
        }
    }

    if( ctrl->ext_tpevent_num )
    {
        for( i = 0; i < ctrl->ext_tpevent_num; i++ )
        {
            layout = &ctrl->ext_tpevent[i];

            if( RECT_T( layout->rect ) <= y && y < RECT_B( layout->rect )
                && RECT_L( layout->rect ) <= x && x < RECT_R( layout->rect ) )
            {
                return layout;
            }
        }
    }

    return NULL;

}




static void RedrawKeytop( KbdControl* ctrl )
{
    int i;
    u16 normal_button_start;
    NNSG2dCharCanvas cc;
	const NNSG2dFont* pFont = KBD_GetFont();
    const int fontHeight = NNS_G2dFontGetHeight(pFont);
    const KbdConfig* const cf = ctrl->conf;

    MI_CpuCopyFast( cf->charimg, sKbdImg, sizeof(sKbdImg) );

    NNS_G2dCharCanvasInitForBG(&cc, sKbdImg, KBD_WIDTH/8, KBD_HEIGHT/8, NNS_G2D_CHARA_COLORMODE_16);

    normal_button_start = getNormalButtonStart( cf );

    //---- writes in center alignment in the center of the KBDButtonRect
    for( i = 0; i < cf->keynum - normal_button_start; i++ )
    {
        const KBDButton* const button = &( cf->layout[normal_button_start + i] );
        const KBDButtonRect* const r = &(button->rect);
        u16 letter = ctrl->keymap[i];
        NNSG2dGlyph glyph;
        int x, y;

        //---- gets glyphs
        NNS_G2dFontGetGlyph(&glyph, pFont, letter);

        //---- calculates display position based on the glyph
        x = r->x + (r->width - glyph.pWidths->glyphWidth) / 2;
        y = r->y + (r->height - fontHeight) / 2;

        //---- draws glyphs
        NNS_G2dCharCanvasDrawGlyph(&cc, pFont, x, y, CHARCOLOR, &glyph);
    }
}




///////////////////////////////////////////////////////////////////////////////
///
/// drawing process (not loaded into GX)
///
///////////////////////////////////////////////////////////////////////////////
static void DrawKeyboard( KbdControl* ctrl )
{
    const KbdConfig *const cf = ctrl->conf;

    // drawing process
    MI_CpuCopyFast( sKbdImg, sIcgBak, sizeof(sKbdImg) );

    if( ctrl->highlight )
    {
        const KBDButton* shiftcaps;

        shiftcaps = & cf->layout[ ctrl->highlight - 1];

        AddBGCharRectangle( (u8*)sIcgBak, KBD_WIDTH, 8,
                            RECT_L( shiftcaps->rect ),
                            RECT_T( shiftcaps->rect ),
                            RECT_R( shiftcaps->rect ),
                            RECT_B( shiftcaps->rect ) );

    }

    // replaces the color of the pressed button
    if( ctrl->pressed
        && ctrl->current
        && ctrl->current->id != ctrl->highlight // no replacement is made if the color has already been replaced on a key like the SHIFT or CAPS LOCK keys
        && ctrl->pressed->id != ctrl->highlight // no replacement is made if the color has already been replaced on a key like the SHIFT or CAPS LOCK keys
        && ctrl->pressed->id != 0 // no color replacement occurs for dummy buttons
        && ctrl->current->id < KBD_BUTTON_ID_EXT_START
        )
    {

        const KBDButton* p = ctrl->pressed; // rotates through the group

        if( ctrl->pressed->id == ctrl->current->id )
        {
            do
            {
                AddBGCharRectangle( (u8*)sIcgBak, KBD_WIDTH, 8,
                                    RECT_L( p->rect ),
                                    RECT_T( p->rect ),
                                    RECT_R( p->rect ),
                                    RECT_B( p->rect ) );
                p += p->group_offset;
            } while( p != ctrl->pressed );
        }
        else
        {
            do
            {
                 // draws four rectangles as below
                 //
                 //         UUUUUUUU
                 //         L......R
                 //         L......R
                 //         L......R
                 //         L......R
                 //         DDDDDDDD

                // U
                AddBGCharRectangle( (u8*)sIcgBak, KBD_WIDTH, 8,
                                    RECT_L( p->rect ),
                                    RECT_T( p->rect ),
                                    RECT_R( p->rect ),
                                    RECT_T( p->rect ) + 1 );
                // D
                AddBGCharRectangle( (u8*)sIcgBak, KBD_WIDTH, 8,
                                    RECT_L( p->rect ) ,
                                    RECT_B( p->rect ) - 1,
                                    RECT_R( p->rect ),
                                    RECT_B( p->rect ));
                // L
                AddBGCharRectangle( (u8*)sIcgBak, KBD_WIDTH, 8,
                                    RECT_L( p->rect ) ,
                                    RECT_T( p->rect ) + 1,
                                    RECT_L( p->rect ) + 1,
                                    RECT_B( p->rect ) - 1);
                // R
                AddBGCharRectangle( (u8*)sIcgBak, KBD_WIDTH, 8,
                                    RECT_R( p->rect ) - 1,
                                    RECT_T( p->rect ) + 1,
                                    RECT_R( p->rect ),
                                    RECT_B( p->rect ) - 1);

                p += p->group_offset;

            } while( p != ctrl->pressed );
        }


    }


}



void KBD_SetDragDropCallback( KBDDragDropCallback callback, void* param )
{
    sControl.dragdropcallback = callback;
    sControl.dragdropparam = param;
}


/*---------------------------------------------------------------------------*

  Returns whether a keyboard button has been pressed at touch-on.

 *---------------------------------------------------------------------------*/
static BOOL isOnPushedButton( const KbdControl* ctrl )
{
    return ( ctrl->pressed && ctrl->current &&
             ctrl->current->id == ctrl->pressed->id );
}



/*---------------------------------------------------------------------------*

  Creates screen data for the software keyboard.

 *---------------------------------------------------------------------------*/
void KBD_FillScreen( GXScrText32x32* scr, ///< [out] screen fill
                      int palette, ///< [in] palette number
                      int kbd_chartop ///< [in] keyboard header character number
    )
{
    int x, y;

    const int WIDTH = KBD_WIDTH / 8; // keyboard character width
    const int HEIGHT = KBD_HEIGHT / 8; // keyboard character height
    const int BLANKCHAR_NO = WIDTH * HEIGHT; // one space follows the keyboard data

    MI_CpuFill16( scr, (u16)BLANKCHAR_NO, sizeof(GXScrText32x32) );

    for( y=0; y<HEIGHT; y++ )
    {
        for( x=0; x<WIDTH; x++ )
        {
            scr->scr[y][x]
                = GX_SCRFMT_TEXT( palette, 0, 0, kbd_chartop + y*WIDTH + x );
        }
    }

    DC_StoreRange( scr, sizeof(GXScrText32x32) );

}




/*---------------------------------------------------------------------------*

  Returns the character code for the character being dragged (0 if there is no code).

 *---------------------------------------------------------------------------*/
u16  KBD_GetDragging( void )
{
    return sControl.dragging;
}


///////////////////////////////////////////////////////////////////////////////
///
/// overall keyboard initialization
///
///////////////////////////////////////////////////////////////////////////////
void KBD_Init( int x, int y, BOOL bootmenu_mode, KBDSoundCallback callback, void* param )
{
    OSOwnerInfo info;

    OS_GetOwnerInfo( &info );

    sPosX = (u16)x;
    sPosY = (u16)y;
    sControl.cursor_input = FALSE;
    sControl.bootmenu_mode = bootmenu_mode;

    //-- sets the sound callback function and its parameters
    sControl.soundcallback = callback;
    sControl.soundparam = param;

    KBD_StartupPadTpInput();  // input initialization

    // process only occurs once at launch
    if( ! s_Initialized )
    {
        s_Initialized = TRUE;

        KBD_InitFont();  // IPL font initialization

        KBD_InitPadRepeat( PAD_PLUS_KEY_MASK | PAD_BUTTON_B | PAD_BUTTON_X );

        // creates conversion mapping for voicing diacritics
        s_dakuten_map[0] = s_handaku_map[0] = s_small_map[0] = s_all_map[0];
        s_dakuten_map[1] = s_all_map[1];
        s_handaku_map[1] = s_all_map[2];
        s_small_map[1]   = s_all_map[3];

        sConfig[ KBD_KIND_ASCII ].keynum = (u16)( RSRC_SIZE( ascii_kbd ) / sizeof(KBDButton) );
        sConfig[ KBD_KIND_JAPANESE ].keynum = (u16)( RSRC_SIZE( jp_kbd ) / sizeof(KBDButton) );
        sConfig[ KBD_KIND_EUROPE ].keynum = (u16)( RSRC_SIZE( picto_kbd ) / sizeof(KBDButton) - 3 );
        sConfig[ KBD_KIND_SIGN   ].keynum = (u16)( RSRC_SIZE( picto_kbd ) / sizeof(KBDButton) );
        sConfig[ KBD_KIND_PICTURE ].keynum = (u16)( RSRC_SIZE( picto_kbd ) / sizeof(KBDButton) );

        sControl.dragdropcallback = NULL;
        sControl.dragdropparam = NULL;

        if( info.language == OS_LANGUAGE_JAPANESE )
        {
            int index = sConfig[ KBD_KIND_JAPANESE ].specialkey_num;

            sControl.cursor = & sConfig[ KBD_KIND_JAPANESE ].layout[ index ] ; // sets the default position for the cursor

            KBD_Change( KBD_KIND_JAPANESE );
        }
        else
        {
            int index = sConfig[ KBD_KIND_ASCII ].specialkey_num;

            sControl.cursor = & sConfig[ KBD_KIND_ASCII ].layout[ index ] ;

            KBD_Change( KBD_KIND_ASCII );

        }

        //-- initializes the pointer to and the number of elements for the additional touch event list
        sControl.ext_tpevent = NULL;
        sControl.ext_tpevent_num = 0;

        //-- clears the additional touch event callback and its parameters
        sControl.ext_tpevent_callback = NULL;
        sControl.ext_tpevent_param = NULL;
    }

}




void KBD_LoadPltt( int palette )
{
    OSOwnerInfo info;
    OS_GetOwnerInfo( &info );

    GXS_LoadBGPltt ( fs_keyboardColor_ncl + sizeof(GXBGPltt16) * info.favoriteColor,
                     sizeof(GXBGPltt16) * palette,  sizeof(GXBGPltt16) );
}


void KBD_Redraw(void)
{
    RedrawKeytop( &sControl );
    DrawKeyboard( &sControl );
    RelocateCursor( &sControl );
}


///////////////////////////////////////////////////////////////////////////////
///
/// keyboard switching
///
///////////////////////////////////////////////////////////////////////////////
void KBD_Change( KBDKind kind )
{
    sControl.kind = kind;

    sControl.pressed = NULL;
    sControl.current = NULL;
    sControl.scancode = 0;
    sControl.input = 0;
    sControl.highlight = 0;
    sControl.dragging = 0;

    //-- when you want to enable changin repeat settings for the touch panel externally
    //-- be aware that repeat settings are initialized here
    KBD_InitAutoRepeatValue( & sControl.repeat, KBD_AUTOREPEAT_FIRST_FRAME, KBD_AUTOREPEAT_NEXT_STEP );

    switch( kind )
    {
    case KBD_KIND_ASCII:    AsciiKBDInit( &sControl );    break;
    case KBD_KIND_JAPANESE: JapaneseKBDInit( &sControl );  break;
    case KBD_KIND_EUROPE:   EuroKBDInit( &sControl ); break;
    case KBD_KIND_SIGN:     SignKBDInit( &sControl ); break;
    case KBD_KIND_PICTURE:  PictoKBDInit( &sControl ); break;
    }

    RedrawKeytop( &sControl );
    DrawKeyboard( &sControl );
    RelocateCursor( &sControl );

}



/*---------------------------------------------------------------------------*

 The keyboard is changed by specifying a number.

 [Japanese]

 [ JAPANESE, ASCII, EUROPE, SIGN, PICTURE ]

 [non-Japanese]

 [ ASCII, EUROPE, JAPANESE, SIGN, PICTURE ]

 *---------------------------------------------------------------------------*/
void KBD_ChangeByNumber( int no )
{
    OSOwnerInfo info;

    KBDKind kinds[] =
        {
            KBD_KIND_ASCII,
            KBD_KIND_EUROPE,
            KBD_KIND_JAPANESE,
            KBD_KIND_SIGN,
            KBD_KIND_PICTURE
        };

    OS_GetOwnerInfo( &info );
    if( info.language == OS_LANGUAGE_JAPANESE )
    {
        kinds[0] = KBD_KIND_JAPANESE;
        kinds[1] = KBD_KIND_ASCII;
        kinds[2] = KBD_KIND_EUROPE;
    }

    KBD_Change( kinds[no] );
    KBDi_SoundCallback( sControl.soundparam, KBD_SE_SWITCH_LETER );

}




/*---------------------------------------------------------------------------*

  Changes to the next keyboard.

 *---------------------------------------------------------------------------*/
void KBD_ChangeToNext( void )
{
    OSOwnerInfo info;
    KBDKind nexts[2][5] =
        {
            { // for Japanese
                KBD_KIND_EUROPE,
                KBD_KIND_ASCII,
                KBD_KIND_SIGN,
                KBD_KIND_PICTURE,
                KBD_KIND_JAPANESE
            },
            { // for non-Japanese
                KBD_KIND_EUROPE,
                KBD_KIND_SIGN,
                KBD_KIND_JAPANESE,
                KBD_KIND_PICTURE,
                KBD_KIND_ASCII
            },
        };

    OS_GetOwnerInfo( &info );
    if( info.language == OS_LANGUAGE_JAPANESE )
    {
        KBD_Change( nexts[0][ KBD_GetKind() ] );
    }
    else
    {
        KBD_Change( nexts[1][ KBD_GetKind() ] );
    }

    KBDi_SoundCallback( sControl.soundparam, KBD_SE_SWITCH_LETER );


}






KBDKind KBD_GetKind( void )
{
    return sControl.kind;
}






///////////////////////////////////////////////////////////////////////////////
///
/// keyboard update process (function that can be called for each loop)
///
///////////////////////////////////////////////////////////////////////////////
void KBD_UpdateValue( u16 paddata, TPData tpdata )
{
    //-- declares a variable for getting the additional touch event state
    u16 pressed_id = KBD_BUTTON_ID_NONE, current_id = KBD_BUTTON_ID_NONE;
    KBDTpEvent ext_tpevent = KBD_TP_NONE;

    KBD_UpdatePadTpInputValue( paddata, tpdata );  // input updates
    KBD_UpdatePadRepeat();  // pad repeat input updates

    sControl.scancode = 0; // clears input
    sControl.input = 0;

    if( sControl.cursor_input )
    {
        sControl.cursor_input = FALSE;
        sControl.scancode = sControl.cursor->id;

        // SE process
        if( sControl.cursor->id ) // dummy button to silence SE
        {
            PlayInputSE( sControl.scancode );
        }

    }
    else if( KBD_GetPadTrigRepeat( ) & PAD_BUTTON_B )
    {
        // disappears with the B button
        sControl.scancode = SCANCODE_DELETE;

        // noise strategy
//        if( ! ( sControl.bootmenu_mode && bm_get_namecomment_length() == 0 ) )
        {
            KBDi_SoundCallback( sControl.soundparam, KBD_SE_BACKSPACE );
        }

    }
    else if( KBD_GetPadTrigRepeat() & PAD_BUTTON_X )
    {
        // shift-state switching
        if ( sControl.kind == KBD_KIND_ASCII )
        {
            if( sControl.highlight == 0 )
            {
                sControl.scancode = SCANCODE_SHIFT;
            }
            else if( sControl.highlight == SCANCODE_SHIFT )
            {
                sControl.scancode = SCANCODE_CAPS;
            }
            else if( sControl.highlight == SCANCODE_CAPS )
            {
                sControl.scancode = SCANCODE_CAPS; // sets to be the same as when the CAPS LOCK is pressed again
            }

        }

//        OS_Printf( "keyinput line : %d\n", __LINE__ );
        KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT );

    }
    else if( KBD_GetPadTrigRepeat( ) & PAD_KEY_UP )
    {
        // cursor movement
        sControl.cursor = MoveCursor( PAD_KEY_UP, sControl.cursor, sControl.conf->layout );
    }
    else if( KBD_GetPadTrigRepeat( ) & PAD_KEY_DOWN )
    {
        sControl.cursor = MoveCursor( PAD_KEY_DOWN, sControl.cursor, sControl.conf->layout );
    }
    else if( KBD_GetPadTrigRepeat( ) & PAD_KEY_LEFT )
    {
        sControl.cursor = MoveCursor( PAD_KEY_LEFT, sControl.cursor, sControl.conf->layout );
    }
    else if( KBD_GetPadTrigRepeat( ) & PAD_KEY_RIGHT )
    {
        sControl.cursor = MoveCursor( PAD_KEY_RIGHT, sControl.cursor, sControl.conf->layout );
    }
    else if( KBD_IsSetPadTrig( PAD_BUTTON_START ) )
    {
        // when starting in boot menu mode, flies to the commit button
        if( sControl.bootmenu_mode )
        {
            sControl.cursor = & s_dummybuttons[0];
        }

    }

    // updates the current button
    if( KBD_IsTouchedTp() )
    {
        sControl.current = SearchButton( &sControl, KBD_GetTpX(), KBD_GetTpY() );

        //-- records the current button ID if the button is being pressed
        if ( sControl.current )
        {
            current_id = sControl.current->id;
        }
    }

    if( KBD_GetTpEvent() == KBD_TP_DOWN ) // touch-on
    {
        // remembers the button at touch-on
        sControl.pressed = SearchButton( &sControl, KBD_GetTpX(), KBD_GetTpY() );

        if( sControl.pressed )
        {

            // SE process
            if( sControl.pressed->id == SCANCODE_DELETE )
            {
                // noise strategy
//                if( ! ( sControl.bootmenu_mode && bm_get_namecomment_length() == 0 ) )
                {

                    KBDi_SoundCallback( sControl.soundparam, KBD_SE_BACKSPACE );
                }
            }
            else if( sControl.pressed->id == SCANCODE_RETURN )
            {
                if( ! sControl.bootmenu_mode ) // no sound occurs in main unit settings
                {
                    KBDi_SoundCallback( sControl.soundparam, KBD_SE_ENTER );
                }
            }
            else if( 0 < sControl.pressed->id && sControl.pressed->id <= sControl.conf->specialkey_num )
            {
                KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT_EX );
            }
            //-- no sound is played for additional touch events
            else if ( sControl.pressed->id >= KBD_BUTTON_ID_EXT_START )
            {
                ext_tpevent = KBD_TP_DOWN;
                pressed_id = sControl.pressed->id;
            }
            else if( sControl.pressed->id )
            {
//        OS_Printf( "keyinput line : %d\n", __LINE__ );

                KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT );
            }

        }

    }
    else if( KBD_GetTpEvent() == KBD_TP_UP ) // touch-off
    {

        // no output at touch-off once auto-repeat is engaged, even once
        if( sControl.pressed && ! KBD_IsRepeatingAutoRepeat( & sControl.repeat ) )
        {
            const KBDButton* button = SearchButton( &sControl, KBD_GetTpX(), KBD_GetTpY() );

            // output if the same key is pressed
            if( button && button->id == sControl.pressed->id )
            {

                sControl.scancode = sControl.pressed->id;

                // SE process
                if( sControl.pressed->id == SCANCODE_DELETE )
                {
                    // noise strategy
//                    if( ! ( sControl.bootmenu_mode && bm_get_namecomment_length() == 0 ) )
                    {
                        KBDi_SoundCallback( sControl.soundparam, KBD_SE_BACKSPACE_REL );
                    }
                }
                else if( sControl.pressed->id == SCANCODE_RETURN )
                {
                    if( ! sControl.bootmenu_mode ) // no sound occurs in main unit settings
                    {
                        KBDi_SoundCallback( sControl.soundparam, KBD_SE_ENTER_REL );
                    }
                }
                else if( 0 < sControl.pressed->id && sControl.pressed->id <= sControl.conf->specialkey_num )
                {
                    KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT_EX_REL );
                }
                else if( sControl.pressed->id )
                {
                    //-- no sound is played for additional touch events
                    if ( sControl.pressed->id < KBD_BUTTON_ID_EXT_START)
                    {
                        KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT_REL );
                    }
                }
            }

        }

        if( sControl.dragdropcallback && sControl.dragging )
        {
            sControl.dragdropcallback( sControl.dragdropparam, sControl.dragging, KBD_GetTpX(), KBD_GetTpY() );
        }

        //-- process to determine and record whether touch-off is associated with an additional touch event
        {
            const KBDButton* button = SearchButton( &sControl, KBD_GetTpX(), KBD_GetTpY() );

            if( button )
            {
                // registers the button prior to release
                current_id  = button->id;
                if ( button->id >= KBD_BUTTON_ID_EXT_START )
                {
                    // if the button prior to release is an additional touch event button,
                    // check that
                    ext_tpevent = KBD_TP_UP;
                }
            }
            else
            {
                // if the location touched before released is undefined
                current_id = KBD_BUTTON_ID_NONE;
            }

            if ( sControl.pressed )
            {
                // register the button pressed first
                pressed_id = sControl.pressed->id;
                if ( sControl.pressed->id >= KBD_BUTTON_ID_EXT_START )
                {
                    // if the button touched first is an additional touch event
                    // button, check that
                    ext_tpevent = KBD_TP_UP;
                }
            }
            else
            {
                // if the location touched first is undefined
                pressed_id = KBD_BUTTON_ID_NONE;
            }
        }

        sControl.dragging = 0;

        sControl.pressed = NULL;
    }


    // repeat process
    if( isOnPushedButton( &sControl ) )
    {
        KBD_UpdateAutoRepeat( &sControl.repeat );

        //-- for additional touch events, record the repeat state if not the touch-on instant
        if ( sControl.pressed->id >= KBD_BUTTON_ID_EXT_START && ext_tpevent != KBD_TP_DOWN )
        {
            // Only this will work for instants exclusive of the repeat trigger.
            // as the pressed_id is not set, it can differentiate from waiting for a repeat trigger
            ext_tpevent = KBD_TP_REPEAT;
        }

        if( KBD_IsTriggerAutoRepeat( &sControl.repeat ) )
        {
            sControl.scancode = sControl.pressed->id;

            // SE process
            if( sControl.pressed->id == SCANCODE_DELETE )
            {
                // noise strategy
//                if( ! ( sControl.bootmenu_mode && bm_get_namecomment_length() == 0 ) )
                {
                    KBDi_SoundCallback( sControl.soundparam, KBD_SE_BACKSPACE );
                }
            }
            else if( sControl.pressed->id == SCANCODE_RETURN )
            {
                if( ! sControl.bootmenu_mode ) // no sound occurs in main unit settings
                {
                    KBDi_SoundCallback( sControl.soundparam, KBD_SE_ENTER );
                }
            }
            else if( 0 < sControl.pressed->id && sControl.pressed->id <= sControl.conf->specialkey_num )
            {
                KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT_EX );
            }
            //-- no sound is played for additional touch events
            //-- and get the event occurrence state
            else if ( sControl.pressed->id >= KBD_BUTTON_ID_EXT_START )
            {
                // set pressed_id to differentiate from a repeat trigger
                pressed_id = sControl.pressed->id;
            }
            else if( sControl.pressed->id )
            {
                KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT );
            }

        }
    }
    else
    {
        KBD_ResetAutoRepeat( &sControl.repeat );
    }

    UpdateDragging( & sControl ); // drag & drop update

    // because the SHIFT status can be changed with a key_func call, update input here
    sControl.input = kbd_get_input_internal( & sControl, sControl.scancode );

    // call the function if the key is pressed
    if( sControl.scancode )
    {
        sControl.conf->key_func( &sControl );
    }


    //-- determines whether an additional event is being dragged
    //-- it's a repeat if the currently pressed button is the same as the button pressed first
    if ( ext_tpevent == KBD_TP_NONE && sControl.pressed
         && sControl.pressed->id >= KBD_BUTTON_ID_EXT_START
         && ( ! sControl.current || sControl.current->id != sControl.pressed->id ) )
    {
        ext_tpevent = KBD_TP_DRAG;
        pressed_id  = sControl.pressed->id;
    }

    //-- if the touch event is an additional touch event,
    //-- call the relevant callback function
    if ( ext_tpevent != KBD_TP_NONE && sControl.ext_tpevent_callback )
    {
        sControl.ext_tpevent_callback( sControl.ext_tpevent_param, ext_tpevent, pressed_id, current_id, KBD_GetTpX(), KBD_GetTpY() );
    }

    DrawKeyboard( &sControl );

}


void KBD_Update( void )
{
    TPData raw_point, disp_point;

    if ( KBD_IsEnableTpSampling() )
	{
        // get touch input
	    while ( TP_RequestRawSampling(&raw_point) != 0 ) {}
    	TP_GetCalibratedPoint( &disp_point, &raw_point );
    }
    else
    {
        // initialize touch input
        disp_point.touch = TP_TOUCH_OFF;
        disp_point.validity = TP_VALIDITY_INVALID_XY;
    }

    KBD_UpdateValue( PAD_Read(), disp_point );
}


u16 KBD_GetInput( void )
{
    return sControl.input;
}



static u16  kbd_get_input_internal( KbdControl* ctrl, u16 scancode )
{
    if( scancode == 0 || scancode >= KBD_BUTTON_ID_EXT_START )
    {
        return 0;
    }
    else if( scancode <= ctrl->conf->specialkey_num )
    {
        switch( scancode )
        {
        case SCANCODE_DELETE: return (u16)'\b';
        case SCANCODE_RETURN: return (u16)'\n';
        case SCANCODE_SPACE:
            if( ctrl->kind == KBD_KIND_JAPANESE )
            {
                return L'@'; // two-byte space
            }
            else
            {
                return L' '; // single-byte space
            }
            break;
        // outputs unique code just for Japanese
        case SCANCODE_DAKUTEN:
            if( ctrl->kind == KBD_KIND_JAPANESE )
            {
                return CHARCODE_DAKUTEN;
            }
            break;
        case SCANCODE_HANDANKU:
            if( ctrl->kind == KBD_KIND_JAPANESE )
            {
                return CHARCODE_HANDAKU;
            }
            break;
        case SCANCODE_SMALL:
            if( ctrl->kind == KBD_KIND_JAPANESE )
            {
                return CHARCODE_SMALL;
            }
            break;
        default:              return 0; // only the above three have returns
        }
    }
    else
    {
        int index = scancode - ctrl->conf->specialkey_num - 1; // as scan codes begin with a 1, subtract 1

		if ( ctrl->keymap[ index ] > 0x20 )
				return ctrl->keymap[ index ];

		return 0;
    }

    SDK_ASSERT( FALSE );
    return 0; // to avoid warning

}



BOOL KBD_IsUsing( void )
{
    return sControl.pressed ? TRUE : FALSE;
}


void* KBD_GetCharImg( void )
{
    DC_StoreRange( sIcgBak, sizeof(sIcgBak) );
    return sIcgBak;
}

u32   KBD_GetCharSize( void )
{
    return sizeof(sIcgBak);
}


void KBD_GetCursorPos( KBDVec2S16 pos[] )
{
    pos[0].x = (s16)( sPosX + RECT_L( sControl.cursor->rect ));   pos[0].y = (s16)(sPosY + RECT_T( sControl.cursor->rect ));
    pos[1].x = (s16)( sPosX + RECT_R( sControl.cursor->rect ));   pos[1].y = (s16)(sPosY + RECT_T( sControl.cursor->rect ));
    pos[2].x = (s16)( sPosX + RECT_L( sControl.cursor->rect ));   pos[2].y = (s16)(sPosY + RECT_B( sControl.cursor->rect ));
    pos[3].x = (s16)( sPosX + RECT_R( sControl.cursor->rect ));   pos[3].y = (s16)(sPosY + RECT_B( sControl.cursor->rect ));
}




// Diacritic conversion. The converted character is returned. If the character cannot be converted, it is returned as is.
// e.g.,  ->   ->   -> 
u16 KBD_ConvertDakuten( u16 c )
{

    if( c == L'@' ) return c; // process to differentiate blank spaces in s_all_map

    // characters with a semi-voiced consonant diacritic are processed after returning normally
    if( KBD_IsHandaku( c ) )
    {
        u16 ret;
        u16 normal = KBD_ConvertHandaku( c );

        ret = ConvertDakutenEtc( s_dakuten_map, normal );

        if( ret != normal ) // if the returned character can be converted, return it
        {
            return ret;
        }
        else // if it cannot be converted, return the original character
        {
            return c;
        }

    }
    else if( KBD_IsSmall( c ) ) // a return attempt is made for small characters as well
    {
        u16 ret;
        u16 normal = KBD_ConvertSmall( c );

        ret = ConvertDakutenEtc( s_dakuten_map, normal );

        if( ret != normal ) // if the returned character can be converted, return it
        {
            return ret;
        }
        else // if it cannot be converted, return the original character
        {
            return c;
        }

    }
    else // convert it as is
    {
        return ConvertDakutenEtc( s_dakuten_map, c );
    }
}


u16 KBD_ConvertHandaku( u16 c )
{

    if( c == L'@' ) return c;

    // voiced consonant diacritic characters are returned normally
    if( KBD_IsDakuten( c ) )
    {
        u16 ret;
        u16 normal = KBD_ConvertDakuten( c );

        ret = ConvertDakutenEtc( s_handaku_map, normal );

        if( ret != normal )
        {
            return ret;
        }
        else
        {
            return c;
        }

    }
    else
    {
        return ConvertDakutenEtc( s_handaku_map, c );
    }
}


u16 KBD_ConvertSmall( u16 c )
{

    if( c == L'@' ) return c;

    // voiced consonant diacritic characters are returned normally
    if( KBD_IsDakuten( c ) )
    {
        u16 ret;
        u16 normal = KBD_ConvertDakuten( c );

        ret = ConvertDakutenEtc( s_small_map, normal );

        if( ret != normal )
        {
            return ret;
        }
        else
        {
            return c;
        }

    }
    else
    {
        return ConvertDakutenEtc( s_small_map, c );
    }

}






/*---------------------------------------------------------------------------*

  The appropriate conversion is performed for voiced and semi-voiced consonant diacritics and for small characters.
  If used in succession, the loop is like  ->  ->  -> .

 *---------------------------------------------------------------------------*/
u16 KBD_ConvertAll( u16 c )
{
    enum{ NORMAL, DAKUTEN, HANDAKU, SMALL, CLASSNUM }; // match to the i value after matching to the s_all_map layout
    int i, offset;

    if( c == L'@' ) return c;


    for( i = 0; i < CLASSNUM; i++ )
    {
        u16* found = wcschr( s_all_map[i], c );

        if( found )
        {
            offset = found - s_all_map[i]; // keep the characters order number
            break;
        }
    }

    if( i < CLASSNUM ) // which did it match?
    {
        int j;

        for( j = 1; j < CLASSNUM; j++ )
        {
            if( s_all_map[ (i+j) % CLASSNUM ][ offset ] != L'@' )
            {
                return s_all_map[ (i+j) % CLASSNUM ][ offset ]; // if a non-blank character is found, return that
            }
        }

    }
    else    // Non-matching characters. E.g., 
    {
        return c;
    }

    return c;

}




BOOL KBD_IsDakuten( u16 c )
{
    return ( wcschr( s_dakuten_map[1], c ) != NULL );
}

BOOL KBD_IsHandaku( u16 c )
{
    return ( wcschr( s_handaku_map[1], c ) != NULL );
}

BOOL KBD_IsSmall( u16 c )
{
    return ( wcschr( s_small_map[1], c ) != NULL );
}


void KBD_InputWithCursor( void )
{
    sControl.cursor_input = TRUE;
}


///////////////////////////////////////////////////////////////////////////////
///
/// Data is overwritten by the @brief grid local.
/// (the grid refers to the 8x8 rectangle one character occupies)
///
///////////////////////////////////////////////////////////////////////////////
static void AddGridRectangle( u8* gridstart,
                              u32 add_value,
                              s32 left, s32 top, s32 right, s32 bottom )
{
    u32 val = 0;
    s32 i;
    u32* grid32;

    //    return; // XXXXXXXXXX

    // it's little-endian, so if you want to use ..ooo..., you'll need 0x000FFF00 data.
    // below follows content for using ..ooo...

    // First place only the width portion in the value. Create 0x00000FFF.

    for ( i=0; i<right-left; i++ )
    {
        val |= add_value << (i<<2);
    }

    val = val << (left<<2);

    grid32 = (u32*)gridstart;

    for ( i=top; i<bottom; ++i )
    {
        u32 masked;

        // transparent portions are not changed to a palette
        masked = val;
        if( ! (grid32[i] & 0xF0000000) ) masked &= 0x0FFFFFFF;
        if( ! (grid32[i] & 0x0F000000) ) masked &= 0xF0FFFFFF;
        if( ! (grid32[i] & 0x00F00000) ) masked &= 0xFF0FFFFF;
        if( ! (grid32[i] & 0x000F0000) ) masked &= 0xFFF0FFFF;
        if( ! (grid32[i] & 0x0000F000) ) masked &= 0xFFFF0FFF;
        if( ! (grid32[i] & 0x00000F00) ) masked &= 0xFFFFF0FF;
        if( ! (grid32[i] & 0x000000F0) ) masked &= 0xFFFFFF0F;
        if( ! (grid32[i] & 0x0000000F) ) masked &= 0xFFFFFFF0;

        grid32[i] += masked;
    }

}




///////////////////////////////////////////////////////////////////////////////
///
/// Data is overwritten for the rectangular space within the @brief text BG character data.
///
///////////////////////////////////////////////////////////////////////////////
static void AddBGCharRectangle( u8* bgchardata, ///< pointer to the character data
                                s32 data_width, ///< character data width [pix]
                                u32 add_value,  ///< number to add
                                s32 left,       ///< upper left X coordinate for the space being overwritten
                                s32 top,        ///< upper left Y coordinate for the space being overwritten
                                s32 right,      ///< bottom right X coordinate for the space being overwritten
                                s32 bottom      ///< bottom right Y coordinate for the space being overwritten
    )
{
    s32 x, y;
    s32 grid_left, grid_top, grid_right, grid_bottom;
    const u32 gridbytenum = 32;

    if( ! ( left < right  ) ) return;
    if( ! ( top  < bottom ) ) return;
    SDK_ASSERT( !( add_value & 0x11111110 ) );

    grid_left = left >> 3;
    grid_top  = top  >> 3;
    grid_right = ((right-1) >> 3) + 1;
    grid_bottom = ((bottom-1) >> 3) + 1;

    for ( y = grid_top; y < grid_bottom; ++y )
    {
        for ( x = grid_left; x < grid_right; ++x )
        {
            u8* gridstart; // The grid's start address. The grid refers to the 8x8 rectangle one character occupies.
            s32 local_left, local_top, local_right, local_bottom; // coordinates within the grid

            local_left   = MAX( 0, left - (x<<3) );
            local_top    = MAX( 0, top  - (y<<3) );
            local_right  = MIN( 8, right - (x<<3) );
            local_bottom = MIN( 8, bottom - (y<<3) );

            gridstart = bgchardata + gridbytenum*( y*(data_width>>3) + x );

            AddGridRectangle( gridstart, add_value,
                              local_left, local_top,
                              local_right, local_bottom );
        }

    }


}




/*---------------------------------------------------------------------------*

  When the keyboard type changes, search for the button in the most
  appropriate place on the next keyboard.

 *---------------------------------------------------------------------------*/
static void RelocateCursor( KbdControl* ctrl )
{
    const KBDButton* nextbutton;
    int x, y;
    enum{ X_OFS = 3, Y_OFS = -3 }; // The center has an environment, making it easier to fail, so it is moved slightly.

    // Moves to the position slightly offset from the button's center and its overlay portion.
    x = sPosX + ( RECT_L( ctrl->cursor->rect ) + RECT_R( ctrl->cursor->rect ) ) / 2 + X_OFS;
    y = sPosY + ( RECT_T( ctrl->cursor->rect ) + RECT_B( ctrl->cursor->rect ) ) / 2 + Y_OFS;

    nextbutton = SearchButton( ctrl, x, y );

    if( nextbutton )
    {
        ctrl->cursor = nextbutton;
    }
    else
    {
        ctrl->cursor = & ctrl->conf->layout[ ctrl->conf->specialkey_num ]; // If it's off, use the default key.
    }

}




/*---------------------------------------------------------------------------*

  Moves the cursor.

 *---------------------------------------------------------------------------*/
static const KBDButton* MoveCursor( int key, const KBDButton* current, const KBDButton* layout )
{
    enum{ SEARCH_STEP = 8, // for each SEARCH_STEP pixel
          SEARCH_NUM_LONG = 10, // Find the SEARCH_NUM number; if not found, end the search. multiple searches for the direction in which the key was pressed.
          SEARCH_NUM_SHORT = 2
        };
    int x_start, x_end, y_start, y_end; // range for searching for the next button
    int x_dir, y_dir;               // search direction
    int x, y;
    const KBDButton* nextbutton;

    SDK_ASSERT( (key & PAD_KEY_UP) || (key & PAD_KEY_DOWN) || ( key & PAD_KEY_LEFT ) || ( key & PAD_KEY_RIGHT ) );
    SDK_ASSERT( current );
    SDK_ASSERT( layout );
    (void)layout;

    KBDi_SoundCallback( sControl.soundparam, KBD_SE_MENU_TP_TAP );

    if( key & PAD_KEY_UP )
    {
        // search up
        x_dir   = - SEARCH_STEP;
        y_dir   = - SEARCH_STEP;

        x_start = sPosX + ( RECT_L( current->rect ) + RECT_R( current->rect ) ) / 2;
        x_end   = x_start + x_dir * SEARCH_NUM_SHORT;
        y_start = sPosY + RECT_T( current->rect );
        y_end   = sPosY + RECT_T( current->rect ) + y_dir * SEARCH_NUM_LONG;
    }
    else if( key & PAD_KEY_DOWN )
    {
        // search down
        x_dir   = + SEARCH_STEP;
        y_dir   = + SEARCH_STEP;

        x_start = sPosX + ( RECT_L( current->rect ) + RECT_R( current->rect ) ) / 2;
        x_end   = x_start + x_dir * SEARCH_NUM_SHORT;
        y_start = sPosY + RECT_B( current->rect );
        y_end   = sPosY + RECT_B( current->rect ) + y_dir * SEARCH_NUM_LONG;
    }
    else if( key & PAD_KEY_LEFT )
    {
        // search left
        x_dir   = - SEARCH_STEP;
        y_dir   = + SEARCH_STEP;

        x_start = sPosX + RECT_L( current->rect );
        x_end   = sPosX + RECT_L( current->rect ) + x_dir * SEARCH_NUM_LONG;
        y_start = sPosY + ( RECT_T( current->rect ) + RECT_B( current->rect ) ) / 2;
        y_end   = y_start + y_dir * SEARCH_NUM_SHORT;
    }
    else if( key & PAD_KEY_RIGHT )
    {
        // search right
        x_dir   = + SEARCH_STEP;
        y_dir   = + SEARCH_STEP;

        x_start = sPosX + RECT_R( current->rect );
        x_end   = sPosX + RECT_R( current->rect ) + x_dir * SEARCH_NUM_LONG;
        y_start = sPosY + ( RECT_T( current->rect ) + RECT_B( current->rect ) ) / 2;
        y_end   = y_start + y_dir * SEARCH_NUM_SHORT;
    }


    for( y = y_start; y != y_end; y += y_dir )
    {
        for( x = x_start; x != x_end; x += x_dir )
        {
            enum{ KBD_W = KBD_WIDTH, KBD_H = KBD_HEIGHT + 16 }; // add the dummy button portion (16)
            int pos_x, pos_y;

            pos_x = x;
            pos_y = y;

            // supplemental for looping
            if( pos_x - sPosX < 0 ) pos_x += KBD_W;
            else if ( pos_x - sPosX > KBD_W ) pos_x -= KBD_W;

            if( pos_y - sPosY < 0 ) pos_y += KBD_H;
            else if ( pos_y - sPosY > KBD_H ) pos_y -= KBD_H;

            nextbutton = SearchButton( & sControl, pos_x, pos_y );

            if( nextbutton && nextbutton != current ) // match, and move if different from current location
            {
                return nextbutton;
            }
        }
    }

    return current; // Don't move if there is no next location.

}





static void AsciiKBDInit( KbdControl* ctrl )
{
    ctrl->conf = &sConfig[ KBD_KIND_ASCII ];
    ctrl->keymap = (u16*)ascii_normal_kbdmap + 1;
}

static void JapaneseKBDInit( KbdControl* ctrl )
{
    ctrl->conf = &sConfig[ KBD_KIND_JAPANESE ];
    ctrl->keymap = (u16*)jp_hiragana_kbdmap + 1 ; // skip the BOM
    ctrl->highlight = SCANCODE_HIRAGANA;
}

static void EuroKBDInit( KbdControl* ctrl )
{
    ctrl->conf = &sConfig[ KBD_KIND_EUROPE ];
    ctrl->keymap = (u16*)euro_normal_kbdmap + 1;
}

static void SignKBDInit( KbdControl* ctrl )
{
    ctrl->conf = & sConfig[ KBD_KIND_SIGN ];
    ctrl->keymap = (u16*)sign_kbdmap + 1;
}

static void PictoKBDInit( KbdControl* ctrl )
{
    ctrl->conf = &sConfig[ KBD_KIND_PICTURE ];
    ctrl->keymap = (u16*)picto_kbdmap + 1;
}







static void JapaneseKBDFunc( KbdControl* ctrl )
{

    if( SCANCODE_HIRAGANA <= ctrl->scancode && ctrl->scancode <= SCANCODE_KATAKANA )
    {
        switch( ctrl->scancode )
        {
        case SCANCODE_HIRAGANA:
            ctrl->keymap = (u16*) jp_hiragana_kbdmap + 1;
            ctrl->highlight = SCANCODE_HIRAGANA;
            break;
        case SCANCODE_KATAKANA:
            ctrl->keymap = (u16*) jp_katakana_kbdmap + 1;
            ctrl->highlight = SCANCODE_KATAKANA;
            break;
        }

        RedrawKeytop( ctrl );

    }

}



static void AsciiKBDFunc( KbdControl* ctrl )
{
    const u8* binaries[] = { ascii_normal_kbdmap,
                             ascii_shift_kbdmap,
                             ascii_caps_kbdmap };

    AsciiEuroCommonKBDFunc( ctrl, binaries );

}



static void EuroKBDFunc( KbdControl* /* ctrl */ )
{
}


static void AsciiEuroCommonKBDFunc( KbdControl * ctrl, const u8* map_binaries[] )
{

    switch( ctrl->highlight )
    {
    case 0: // normal state
        if( ctrl->scancode == SCANCODE_SHIFT )
        {
            ctrl->highlight = SCANCODE_SHIFT;
        }
        else if( ctrl->scancode == SCANCODE_CAPS )
        {
            ctrl->highlight = SCANCODE_CAPS;
        }
        break;
    case SCANCODE_SHIFT: // shift state
        if( ctrl->scancode == SCANCODE_SHIFT )
        {
            ctrl->highlight = 0;
        }
        else if( ctrl->scancode == SCANCODE_CAPS )
        {
            ctrl->highlight = SCANCODE_CAPS;
        }
        else
        {
            ctrl->highlight = 0;
        }
        break;
    case SCANCODE_CAPS: // caps state
        if( ctrl->scancode == SCANCODE_SHIFT )
        {
            ctrl->highlight = SCANCODE_SHIFT;
        }
        else if( ctrl->scancode == SCANCODE_CAPS )
        {
            ctrl->highlight = 0;
        }
        break;
    }

    switch( ctrl->highlight )
    {
    case 0:     ctrl->keymap = (u16*) map_binaries[0] + 1; break;
    case SCANCODE_SHIFT: ctrl->keymap = (u16*) map_binaries[1] + 1; break;
    case SCANCODE_CAPS:  ctrl->keymap = (u16*) map_binaries[2] + 1; break;
    }

    RedrawKeytop( ctrl );

}


static void SignKBDFunc( KbdControl* /* ctrl */ )
{
}

static void PictoKBDFunc( KbdControl* /* ctrl */ )
{
}




/*---------------------------------------------------------------------------*

  Mutually converts voiced and semi-voiced consonant diacritic characters.

 *---------------------------------------------------------------------------*/
static u16 ConvertDakutenEtc( const u16* map[], u16 c  )
{
    int i;
    int offset;
    u16* p;

    for( i = 0; i < 2; i++ )
    {
        p = wcschr( map[i], c );

        if( p )
        {
            offset = p - map[i];

            if( map[ 1 - i ][ offset ] != L'@' )
            {
                return map[ 1 - i ][ offset ];
            }
            else
            {
                return c;
            }

        }

    }

    return c;

}



static void PlayInputSE( u16 scancode )
{
    if( scancode == SCANCODE_DELETE )
    {
//        if( ! ( sControl.bootmenu_mode && bm_get_namecomment_length() == 0 ) )
        {
            KBDi_SoundCallback( sControl.soundparam, KBD_SE_BACKSPACE );
        }
    }
    else if( scancode == SCANCODE_RETURN )
    {
        if( sControl.bootmenu_mode ) // sound is played in main unit settings
        {
            KBDi_SoundCallback( sControl.soundparam, KBD_SE_ERROR );
        }
        else
        {
            KBDi_SoundCallback( sControl.soundparam, KBD_SE_ENTER );
        }
    }
    else if( 0 < scancode && scancode <= sControl.conf->specialkey_num )
    {
        KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT_EX );
    }
    else if( scancode )
    {
//        OS_Printf( "keyinput line : %d\n", __LINE__ );
        KBDi_SoundCallback( sControl.soundparam, KBD_SE_KEYINPUT );
    }
}



/*---------------------------------------------------------------------------*

  Updates the object being dragged.

 *---------------------------------------------------------------------------*/
static void UpdateDragging( KbdControl* ctrl )
{
    const KBDButton* button;

    // determines which key you are above
    button = SearchButton( (const KbdControl*)ctrl, KBD_GetTpX(), KBD_GetTpY() );

    if( ctrl->pressed )
    {
        // If the button is pressed, and is above a location other than that of the button, it is being dragged.
        if( ! button || ctrl->pressed != button )
        {
            u16 code = kbd_get_input_internal( ctrl, ctrl->pressed->id );

            if( code >= L' ' ) // special codes are excepted
            {
                ctrl->dragging = code;
            }

        }

    }


}


/*---------------------------------------------------------------------------*

  Sets the additional touch events.

 *---------------------------------------------------------------------------*/
void KBD_SetExtTpEvent( const KBDButton event_list[], u16 num_element, KBDExtTpEventCallback callback, void* param )
{
    u16 i;

    sControl.ext_tpevent = event_list;
    sControl.ext_tpevent_num = num_element;
    sControl.ext_tpevent_callback = callback;
    sControl.ext_tpevent_param = param;

    // checks the group offset and button ID for the additional touch event structure
    for ( i = 0; i < sControl.ext_tpevent_num; i++ )
    {
        // the group offset is always 0
        SDK_ASSERT( ! sControl.ext_tpevent[i].group_offset );
        // the button ID must be in the range from 0x0100 to 0xfeff
        SDK_ASSERT( sControl.ext_tpevent[i].id >= KBD_BUTTON_ID_EXT_START
                    && sControl.ext_tpevent[i].id <= KBD_BUTTON_ID_EXT_END );
    }
}


/*---------------------------------------------------------------------------*

  Calls the sound callback.

 *---------------------------------------------------------------------------*/
static void KBDi_SoundCallback( void* param, int sound )
{

    if (sControl.soundcallback)
    {
        sControl.soundcallback( param, sound );
    }
}

