/*---------------------------------------------------------------------------*
  Project:  SoftKeyboard
  File:     kbd_autoRepeat.c

  Copyright 2005 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: kbd_autoRepeat.c,v $
  Revision 1.2  2005/07/04 01:10:58  hatamoto_minoru
  comment revisions

  Revision 1.1  2005/06/14 11:30:05  hatamoto_minoru
  Initial registration

  $NoKeywords: $
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
  processes related to auto-repeat
 *---------------------------------------------------------------------------*/
#include <nitro.h>

#include "kbd_autoRepeat.h"
#include "kbd_input.h"

typedef struct
{
    KBDAutoRepeat ar;
    u16         mask; // which buttons are repeated
    u16    repeating; // repeating buttons
} tPadRepeat;


static tPadRepeat s_pad_repeat;


// static functions


/**--------------------------------------------------------------------------
  The auto-repeat data structues are initialized with default values.

  arguments: repeat : pointer to the initialized auto-repeat structure
  ---------------------------------------------------------------------------*/
void KBD_InitAutoRepeat( KBDAutoRepeat* repeat )
{
    repeat->frame = 0;
    repeat->is_repeating = FALSE;

    repeat->first_frame = KBD_AUTOREPEAT_FIRST_FRAME;
    repeat->next_frame  = KBD_AUTOREPEAT_FIRST_FRAME + KBD_AUTOREPEAT_NEXT_STEP;
}

/**--------------------------------------------------------------------------
  The auto-repeat data structues are initialized with repeat parameters.

  arguments: repeat : pointer to the initialized auto-repeat structure
         first  : the number of counters until the first trigger
         next   : the number of counters until the next trigger
  ---------------------------------------------------------------------------*/
void KBD_InitAutoRepeatValue( KBDAutoRepeat* repeat, int first, int step )
{
    repeat->frame = 0;
    repeat->is_repeating = FALSE;

    repeat->first_frame = (u16)first;
    repeat->next_frame  = (u16)(first + step);
}

/**--------------------------------------------------------------------------
  Increases the auto-repeat structure counter by one.

  arguments: repeat : pointer to the initialized auto-repeat structure
  ---------------------------------------------------------------------------*/
void KBD_UpdateAutoRepeat( KBDAutoRepeat* repeat )
{

    repeat->frame++;

    // sets a flag for repeat status
    if ( repeat->frame == repeat->first_frame ||
         repeat->frame == repeat->next_frame )
    {
        repeat->is_repeating = TRUE;
    }


    if ( repeat->frame > repeat->next_frame )
    {
        // If first_frame is entered, it will cause a trigger, so the next frame is used.
        repeat->frame = (u16)(repeat->first_frame + 1);
    }
}

/**--------------------------------------------------------------------------
  Resets the auto-repeat structure.

  arguments: repeat : pointer to the initialized auto-repeat structure
  ---------------------------------------------------------------------------*/
void KBD_ResetAutoRepeat( KBDAutoRepeat* repeat )
{
    repeat->frame = 0;
    repeat->is_repeating = FALSE;
}

/**--------------------------------------------------------------------------
  Tests whether a trigger has been caused for the specified auto-repeat structure.

  arguments: repeat : pointer to the initialized auto-repeat structure
  ---------------------------------------------------------------------------*/
BOOL KBD_IsTriggerAutoRepeat( KBDAutoRepeat* repeat )
{
    if ( repeat->frame == repeat->first_frame ||
         repeat->frame == repeat->next_frame )
    {
        return TRUE;
    }

    return FALSE;
}




/*---------------------------------------------------------------------------*

  Returns a value indicating whether repeating or not.

 *---------------------------------------------------------------------------*/
BOOL KBD_IsRepeatingAutoRepeat( const KBDAutoRepeat* repeat )
{
    return repeat->is_repeating;
}

void KBD_InitPadRepeat( u16 mask )
{
    KBD_InitAutoRepeat( & s_pad_repeat.ar );
    s_pad_repeat.mask = mask;
    s_pad_repeat.repeating = 0;
}

void KBD_InitPadRepeatValue( u16 mask, int first, int step )
{
    KBD_InitAutoRepeatValue( & s_pad_repeat.ar, first, step );
    s_pad_repeat.mask = mask;
    s_pad_repeat.repeating = 0;
}


void KBD_UpdatePadRepeat( void )
{

    if( ! s_pad_repeat.repeating && KBD_IsSetPadTrig( s_pad_repeat.mask ) ) // if not repeating, a repeat check is triggered
    {
        s_pad_repeat.repeating = (u16)KBD_IsSetPadTrig( s_pad_repeat.mask );
    }
    else if( KBD_IsSetPadPress( s_pad_repeat.repeating ) ) // repeat update is the repeating key is being pressed
    {
        KBD_UpdateAutoRepeat( & s_pad_repeat.ar );
    }
    else // reset if not
    {
        KBD_ResetAutoRepeat( & s_pad_repeat.ar );
        s_pad_repeat.repeating = 0;
    }

}


u16 KBD_GetPadRepeat( void )
{
    return (u16)( KBD_IsTriggerAutoRepeat( & s_pad_repeat.ar ) ? KBD_IsSetPadPress( s_pad_repeat.repeating ) : 0 );
}


// Gets the trigger and repeat. The trigger can be got without the repeat mask.
u16 KBD_GetPadTrigRepeat( void )
{
    if( KBD_IsSetPadTrig( PAD_ALL_MASK ) )
    {
        return (u16)KBD_IsSetPadTrig( PAD_ALL_MASK );
    }
    else
    {
        return KBD_GetPadRepeat();
    }

}


//-- Gets the pad repeat frame.
void KBD_GetPadRepeatFrame( int* first, int* step )
{

    *first = s_pad_repeat.ar.first_frame;
    *step  = s_pad_repeat.ar.next_frame - s_pad_repeat.ar.first_frame;
}

//-- Gets the pad mask value.
u16 KBD_GetPadRepeatMask( void )
{

    return s_pad_repeat.mask;
}

