/*---------------------------------------------------------------------------*
  Project:  SoftKeyboard
  File:     kbd_input.c

  Copyright 2005 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: kbd_input.c,v $
  Revision 1.2  2005/07/04 01:10:58  hatamoto_minoru
  comment revisions

  Revision 1.1  2005/06/14 11:30:05  hatamoto_minoru
  initial registration

  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include "kbd_input.h"
#include "kbd_util.h"

// touch panel structure
typedef struct
{
	KBDVec2S16;
	u32 rep_cnt;
	KBDTpEvent event;
	u16 touch;
    u16 pad;
} KBDTpData;

// input structure
typedef struct
{
	struct
	{
		u32 rep_cnt; // count for presses of key-mask objects
		u16 trig, press, release, key_msk;
	} pad;

	struct
	{
		TPData   old;      // an object from the prior
		KBDTpData cur;     // current
		BOOL     sampling; // is sampling performed?
	} tp;
} SUInput;

static SUInput s_input;

/* KBD_ClearPadTpInput ----------------------------------------------------- *

	All input is cleared.
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_ClearPadTpInput(void)
{
	MI_CpuClearFast(&s_input, sizeof(SUInput));
}

/* KBD_StartupPadTpInput --------------------------------------------------- *

	All input is cleared.
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_StartupPadTpInput(void)
{
	KBD_ClearPadTpInput();
	s_input.tp.sampling = TRUE;
}

/* KBD_UpdatePadTpInputValue ----------------------------------------------- *

	Performs update process for input.
	(gets pad or touch panel input externally)
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_UpdatePadTpInputValue(u16 paddata, TPData tpdata)
{
	// key
	{
		s_input.pad.trig = (u16) (paddata & (paddata ^ s_input.pad.press));
		s_input.pad.release = (u16) ((~paddata) & (paddata ^ s_input.pad.press));
		s_input.pad.press = paddata;

		if (s_input.pad.press & s_input.pad.key_msk) { ++ s_input.pad.rep_cnt;}
		else                                         { s_input.pad.rep_cnt = 0;}
	}

	// touch
	if (s_input.tp.sampling)
	{
		// use the value from the prior frame if not valid
		if (tpdata.validity != TP_VALIDITY_VALID)
		{
			tpdata = s_input.tp.old;
			tpdata.touch = 0;
		}

		// event generation
		if (tpdata.touch != s_input.tp.old.touch)
		{
			s_input.tp.cur.event = tpdata.touch ? KBD_TP_DOWN : KBD_TP_UP;
			s_input.tp.cur.rep_cnt = 0;
		}
		else if (tpdata.touch)
		{
			s_input.tp.cur.event = KBD_TP_REPEAT;
			++ s_input.tp.cur.rep_cnt;
		}
		else { s_input.tp.cur.event = KBD_TP_NONE;}

		// Update coordinates when pressed. Do nothing otherwise.
		// Store the last touched location.
		if (tpdata.touch)
		{
			s_input.tp.cur.x = (s16)tpdata.x;
			s_input.tp.cur.y = (s16)tpdata.y;
		}

		// remainder
		s_input.tp.cur.touch = tpdata.touch;
		s_input.tp.old = tpdata;
	}
}

/* KBD_UpdatePadTpInput ---------------------------------------------------- *

	Performs update process for input.
	(gets pad or touch panel data internally)
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_UpdatePadTpInput(void)
{
    TPData raw_point, disp_point;

    if (s_input.tp.sampling)
	{
        // get touch input
	    while (TP_RequestRawSampling(&raw_point) != 0) {}
    	TP_GetCalibratedPoint(&disp_point, &raw_point);
    }
    else {
        // initialize touch input
        disp_point.touch = TP_TOUCH_OFF;
        disp_point.validity = TP_VALIDITY_INVALID_XY;
    }

    KBD_UpdatePadTpInputValue(PAD_Read(), disp_point);
}

/* KBD_IsSetPadTrig -------------------------------------------------------- *

	Detects pad trigger.
	key     relevant key
	returned value: BOOL
 * ------------------------------------------------------------------------- */
BOOL KBD_IsSetPadTrig(int key)
{
	return s_input.pad.trig & key;
}

/* KBD_IsSetPadPress ------------------------------------------------------- *

	Detects pad press.
	key     relevant key
	returned value: BOOL
 * ------------------------------------------------------------------------- */
BOOL KBD_IsSetPadPress(int key)
{
	return s_input.pad.press & key;
}

/* KBD_IsSetPadRelease ----------------------------------------------------- *

	Detects pad release.
	key     relevant key
	returned value: BOOL
 * ------------------------------------------------------------------------- */
BOOL KBD_IsSetPadRelease(int key)
{
	return s_input.pad.release & key;
}

/* KBD_IsTouchedTp --------------------------------------------------------- *

	Determines if the keyboard has been touched or not.
	returned value: BOOL
 * ------------------------------------------------------------------------- */
BOOL KBD_IsTouchedTp(void)
{
	return s_input.tp.cur.touch;
}

/* KBD_GetTpEvent ---------------------------------------------------------- *

	Returns an event.
	returned value: event
 * ------------------------------------------------------------------------- */
KBDTpEvent KBD_GetTpEvent(void)
{
	return s_input.tp.cur.event;
}

/* KBD_GetTpX -------------------------------------------------------------- *

	Returns the X for the touched location.
	returned value: X
 * ------------------------------------------------------------------------- */
u16 KBD_GetTpX(void)
{
	return (u16)s_input.tp.cur.x;
}

/* KBD_GetTpY -------------------------------------------------------------- *

	Returns the Y for the touched location.
	returned value: Y
 * ------------------------------------------------------------------------- */
u16 KBD_GetTpY(void)
{
	return (u16)s_input.tp.cur.y;
}


#if 0
/* KBD_GetTpRepeatFrame ---------------------------------------------------- *

	number of repeat frames
	returned value: number of repeat frames
 * ------------------------------------------------------------------------- */
u32 KBD_GetTpRepeatFrame(void)
{
	return s_input.tp.cur.rep_cnt;
}
#endif

#if 0
/* KBD_SetPadRepeatMask ---------------------------------------------------- *

	Sets the key repeat mask.
	msk     mask
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_SetPadRepeatMask(u16 msk)
{
	s_input.pad.key_msk = msk;
}

/* KBD_GetPadRepeatFrame --------------------------------------------------- *

	Returns the number of key repeat frames.
	returned value: number of key repeat frames
 * ------------------------------------------------------------------------- */
u32 KBD_GetPadRepeatFrame(void)
{
	return s_input.pad.rep_cnt;
}
#endif


/* KBD_EnableTpSampling ---------------------------------------------------- *

	Performs sampling.
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_EnableTpSampling(void)
{
	s_input.tp.sampling = TRUE;
}

/* KBD_DisableTpSampling --------------------------------------------------- *

	Sampling not performed.
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_DisableTpSampling(void)
{
	s_input.tp.sampling = FALSE;
}

/* KBD_IsEnableTpSampling -------------------------------------------------- *

	Checks sampling permissions.
	returned value: sampling permission
 * ------------------------------------------------------------------------- */
BOOL KBD_IsEnableTpSampling(void)
{
	return s_input.tp.sampling;
}

/* KBD_ClearPadInput ------------------------------------------------------- *

	Clears the input from the key specified by the mask.
	mask    relevant key mask
	returned value: none
 * ------------------------------------------------------------------------- */
void KBD_ClearPadInput(int mask)
{
	u16 tmp = (u16) mask;
	s_input.pad.press   &= (~tmp);
	s_input.pad.trig    &= (~tmp);
	s_input.pad.release &= (~tmp);
}

